import logging
import string

from termcolor import cprint

from .code_change import CodeChange, CodeChangeAction
from .code_change_display import get_added_block, get_removed_block
from .user_input_manager import UserInputManager


def resolve_insertion_conflicts(
    changes: list[CodeChange], user_input_manager: UserInputManager, code_file_manager
) -> list[CodeChange]:
    """merges insertion conflicts into one singular code change"""
    insert_changes = list(
        filter(
            lambda change: change.action == CodeChangeAction.Insert,
            sorted(changes, reverse=True),
        )
    )
    new_insert_changes = []
    cur = 0
    while cur < len(insert_changes):
        end = cur + 1
        while (
            end < len(insert_changes)
            and insert_changes[end].first_changed_line
            == insert_changes[cur].first_changed_line
        ):
            end += 1
        if end > cur + 1:
            logging.debug("insertion conflict")
            cprint("Insertion conflict:", "red")
            for i in range(end - cur):
                cprint(f"({string.printable[i]})", "green")
                cprint("\n".join(insert_changes[cur + i].code_lines), "light_cyan")
            cprint(
                "Type the order in which to insert changes (omit for no preference):"
            )
            user_input = user_input_manager.collect_user_input()
            new_code_lines = []
            used = set()
            for c in user_input:
                index = string.printable.index(c) if c in string.printable else -1
                if index < end - cur and index != -1:
                    new_code_lines += insert_changes[cur + index].code_lines
                    used.add(index)
            for i in range(end - cur):
                if i not in used:
                    new_code_lines += insert_changes[cur + i].code_lines
            new_change = CodeChange(
                insert_changes[cur].json_data,
                new_code_lines,
                insert_changes[cur].git_root,
                code_file_manager,
            )
            new_insert_changes.append(new_change)
        else:
            new_insert_changes.append(insert_changes[cur])
        cur = end
    return sorted(
        list(filter(lambda change: change.action != CodeChangeAction.Insert, changes))
        + new_insert_changes,
        reverse=True,
    )


def resolve_non_insertion_conflicts(
    changes: list[CodeChange], user_input_manager: UserInputManager
) -> list[CodeChange]:
    """resolves delete-replace conflicts and asks user on delete-insert or replace-insert conflicts"""
    min_changed_line = changes[0].last_changed_line + 1
    removed_changes = set()
    for i, change in enumerate(changes):
        if change.last_changed_line >= min_changed_line:
            if change.action == CodeChangeAction.Insert:
                logging.debug("insertion inside removed block")
                if changes[i - 1].action == CodeChangeAction.Delete:
                    keep = True
                else:
                    cprint(
                        "\nInsertion conflict: Lines inserted inside replaced block\n",
                        "light_red",
                    )
                    print(get_removed_block(changes[i - 1]))
                    print(get_added_block(change, prefix=">", color=None))
                    print(get_added_block(changes[i - 1]))
                    cprint("Keep this insertion?")
                    keep = user_input_manager.ask_yes_no(default_yes=True)
                if keep:
                    change.first_changed_line = changes[i - 1].first_changed_line - 0.5
                    change.last_changed_line = change.first_changed_line
                else:
                    removed_changes.add(i)

            else:
                change.last_changed_line = min_changed_line - 1
                change.first_changed_line = min(
                    change.first_changed_line, changes[i - 1].first_changed_line
                )
        min_changed_line = change.first_changed_line
    return [change for i, change in enumerate(changes) if i not in removed_changes]



import math

from pygments import highlight
from pygments.formatters import TerminalFormatter
from termcolor import colored

from .code_change import CodeChangeAction

change_delimiter = 60 * "="


def _remove_extra_empty_lines(lines):
    if not lines:
        return []

    # Find the first non-empty line
    start = 0
    while start < len(lines) and not lines[start].strip():
        start += 1

    # Find the last non-empty line
    end = len(lines) - 1
    while end > start and not lines[end].strip():
        end -= 1

    # If all lines are empty, keep only one empty line
    if start == len(lines):
        return [" "]

    # Return the list with only a maximum of one empty line on either side
    return lines[max(start - 1, 0) : end + 2]


def _prefixed_lines(code_change, lines, prefix):
    return "\n".join(
        [
            prefix
            + " " * (code_change.line_number_buffer - len(prefix))
            + line.strip("\n")
            for line in lines
        ]
    )


def print_change(code_change):
    to_print = [
        get_file_name(code_change),
        change_delimiter,
        get_previous_lines(code_change),
        get_removed_block(code_change),
        get_added_block(code_change),
        get_later_lines(code_change),
        change_delimiter,
    ]
    for s in to_print:
        if s:
            print(s)


def get_file_name(code_change):
    file_name = code_change.file
    action = code_change.action
    color = (
        "light_red"
        if action == CodeChangeAction.DeleteFile
        else ("light_green" if action == CodeChangeAction.CreateFile else "light_blue")
    )
    return colored(
        f"\n{file_name}{'*' if action == CodeChangeAction.CreateFile else ''}",
        color=color,
    )


def get_removed_block(code_change, prefix="-", color="red"):
    if code_change.action.has_removals():
        if code_change.action == CodeChangeAction.DeleteFile:
            changed_lines = code_change.file_lines
        else:
            changed_lines = code_change.file_lines[
                code_change.first_changed_line - 1 : code_change.last_changed_line
            ]

        removed = _prefixed_lines(code_change, changed_lines, prefix)
        if removed:
            return colored(removed, color=color)
    return ""


def get_added_block(code_change, prefix="+", color="green"):
    if code_change.action.has_additions():
        added = _prefixed_lines(code_change, code_change.code_lines, prefix)
        if added:
            return colored(added, color=color)
    return ""


def get_previous_lines(code_change, num=2):
    if not code_change.action.has_surrounding_lines():
        return ""
    lines = _remove_extra_empty_lines(
        [
            code_change.file_lines[i]
            for i in range(
                max(0, math.ceil(code_change.first_changed_line) - (num + 1)),
                min(
                    math.ceil(code_change.first_changed_line) - 1,
                    len(code_change.file_lines),
                ),
            )
        ]
    )
    numbered = [
        (str(math.ceil(code_change.first_changed_line) - len(lines) + i) + ":").ljust(
            code_change.line_number_buffer
        )
        + line
        for i, line in enumerate(lines)
    ]

    prev = "\n".join(numbered)
    if prev:
        h_prev = highlight(prev, code_change.lexer, TerminalFormatter(bg="dark"))
        return h_prev
    return ""


def get_later_lines(code_change, num=2):
    if not code_change.action.has_surrounding_lines():
        return ""
    lines = _remove_extra_empty_lines(
        [
            code_change.file_lines[i]
            for i in range(
                max(0, int(code_change.last_changed_line)),
                min(
                    int(code_change.last_changed_line) + num,
                    len(code_change.file_lines),
                ),
            )
        ]
    )
    numbered = [
        (str(int(code_change.last_changed_line) + 1 + i) + ":").ljust(
            code_change.line_number_buffer
        )
        + line
        for i, line in enumerate(lines)
    ]

    later = "\n".join(numbered)
    if later:
        h_later = highlight(later, code_change.lexer, TerminalFormatter(bg="dark"))
        return h_later
    return ""



from .code_change import CodeChange
from .code_file_manager import CodeFileManager
from .config_manager import ConfigManager
from .llm_api import CostTracker, check_model_availability, choose_model, count_tokens
from .parsing import run_async_stream_and_parse_llm_response
from .prompts import system_prompt


class Conversation:
    def __init__(self, config: ConfigManager, cost_tracker: CostTracker):
        self.messages = []
        self.add_system_message(system_prompt)
        self.cost_tracker = cost_tracker
        self.allow_32k = check_model_availability(config.allow_32k())

    def add_system_message(self, message: str):
        self.messages.append({"role": "system", "content": message})

    def add_user_message(self, message: str):
        self.messages.append({"role": "user", "content": message})

    def add_assistant_message(self, message: str):
        self.messages.append({"role": "assistant", "content": message})

    def get_model_response(
        self, code_file_manager: CodeFileManager, config: ConfigManager
    ) -> (str, list[CodeChange]):
        messages = self.messages.copy()

        code_message = code_file_manager.get_code_message()
        messages.append({"role": "system", "content": code_message})

        model, num_prompt_tokens = choose_model(messages, self.allow_32k)

        state = run_async_stream_and_parse_llm_response(
            messages, model, code_file_manager
        )

        self.cost_tracker.display_api_call_stats(
            num_prompt_tokens,
            count_tokens(state.message),
            model,
            state.time_elapsed,
        )

        self.add_assistant_message(state.message)
        return state.explanation, state.code_changes



import os
from textwrap import dedent

from mentat.app import run


def test_insert(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary file
                        # with 2 lines"""))

    mock_collect_user_input.side_effect = [
        "Insert a comment between both lines",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will insert a comment between both lines.

        Steps: 1. Insert a comment after the first line

        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # This is a temporary file
                        # I inserted this comment
                        # with 2 lines""")
    assert content == expected_content


def test_replace(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary file
                        # with 2 lines"""))

    mock_collect_user_input.side_effect = [
        "Replace both lines with one comment",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will replace both lines with one comment

        Steps: 1. Replace both lines with one comment

        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # I inserted this comment""")
    assert content == expected_content


def test_delete(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary
                        # file
                        # with 4 
                        # lines"""))

    mock_collect_user_input.side_effect = [
        "Delete the middle two lines",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will delete the middle two lines

        Steps: 1. Delete the middle two lines

        @@start
        {{
            "file": "{file_name}",
            "action": "delete",
            "start-line": 2,
            "end-line": 3
        }}
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # This is a temporary
                        # lines""")
    assert content == expected_content


def test_create_file(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    mock_collect_user_input.side_effect = [
        "Create a new file called temp.py",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will create a new file called temp.py

        Steps: 1. Create a new file called temp.py

        @@start
        {{
            "file": "{file_name}",
            "action": "create-file"
        }}
        @@code
        # I created this file
        @@end""".format(file_name=temp_file_name))])

    run(["."])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # I created this file""")
    print(content)
    print(expected_content)
    assert content == expected_content


def test_delete_file(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "incredibly_temp.py"
    with open(temp_file_name, "w") as f:
        f.write("# I am not long for this world")

    mock_collect_user_input.side_effect = [
        "Delete the file",
        "y",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will delete the file

        Steps: 1. Delete the file

        @@start
        {{
            "file": "{file_name}",
            "action": "delete-file"
        }}
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    assert not os.path.exists(temp_file_name)


def test_multiple_blocks(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary file
                        # with 4 lines
                        # just for
                        # good measure"""))

    mock_collect_user_input.side_effect = [
        (
            "Insert a comment between the first 2 lines and then replace the last line"
            " with 'better measure'"
        ),
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will insert a comment between the first two lines
        and then replace the last line with 'better measure'

        Steps: 1. Insert a comment
               2. Replace last line

        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 4,
            "end-line": 4
        }}
        @@code
        # better measure
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # This is a temporary file
                        # I inserted this comment
                        # with 4 lines
                        # just for
                        # better measure""")
    assert content == expected_content



import json
import logging
import os
from importlib import resources
from pathlib import Path

mentat_dir_path = os.path.join(Path.home(), ".mentat")

# package_name should always be "mentat" - but this will work if package name is changed
package_name = __name__.split(".")[0]
config_file_name = "default_config.json"
user_config_file_name = "config.json"
user_config_path = os.path.join(mentat_dir_path, user_config_file_name)


class ConfigManager:
    def __init__(self):
        if os.path.exists(user_config_path):
            with open(user_config_path) as config_file:
                self.user_config = json.load(config_file)
        else:
            self.user_config = {}
        with resources.files(package_name).joinpath(config_file_name).open(
            "r"
        ) as config_file:
            self.default_config = json.load(config_file)

    def input_style(self) -> list[list[str]]:
        return self._get_key("input-style")

    def allow_32k(self) -> bool:
        return self._get_key("allow-32k")

    def filetype_include_list(self) -> list[str]:
        return self._get_key("filetype-include-list")

    def filetype_exclude_list(self) -> list[str]:
        return self._get_key("filetype-exclude-list")

    def file_exclude_glob_list(self) -> list[str]:
        return self._get_key("file-exclude-glob-list")

    def _get_key(self, key: str):
        if key in self.user_config:
            return self.user_config[key]
        elif key in self.default_config:
            return self.default_config[key]
        else:
            logging.warning(f"No value for config key {key} found")
            return None



import argparse
import logging
import os
from typing import Iterable

from termcolor import cprint

from .code_change import CodeChange, CodeChangeAction
from .code_change_display import print_change
from .code_file_manager import CodeFileManager
from .config_manager import ConfigManager, mentat_dir_path
from .conversation import Conversation
from .llm_api import CostTracker, count_tokens, setup_api_key
from .logging_config import setup_logging
from .user_input_manager import UserInputManager


def run_cli():
    parser = argparse.ArgumentParser(
        description="Run conversation with command line args"
    )
    parser.add_argument("paths", nargs="*", help="Paths to directories or files")
    paths = parser.parse_args().paths
    run(paths)


def run(paths: Iterable[str]):
    os.makedirs(mentat_dir_path, exist_ok=True)
    setup_logging()
    setup_api_key()
    logging.debug(f"Paths: {paths}")

    cost_tracker = CostTracker()
    try:
        loop(paths, cost_tracker)
    except (EOFError, KeyboardInterrupt) as e:
        print(e)
    finally:
        cost_tracker.display_total_cost()


def loop(paths: Iterable[str], cost_tracker: CostTracker) -> None:
    config = ConfigManager()
    conv = Conversation(config, cost_tracker)
    user_input_manager = UserInputManager(config)
    code_file_manager = CodeFileManager(paths, user_input_manager, config)

    tokens = count_tokens(code_file_manager.get_code_message())
    cprint(f"\nFile token count: {tokens}", "cyan")
    cprint("Type 'q' or use Ctrl-C to quit at any time.\n", color="cyan")
    cprint("What can I do for you?", color="light_blue")
    need_user_request = True
    while True:
        if need_user_request:
            user_response = user_input_manager.collect_user_input()
            conv.add_user_message(user_response)
        explanation, code_changes = conv.get_model_response(code_file_manager, config)
        warn_user_wrong_files(code_file_manager, code_changes)

        if code_changes:
            need_user_request = get_user_feedback_on_changes(
                config, conv, user_input_manager, code_file_manager, code_changes
            )
        else:
            need_user_request = True


def warn_user_wrong_files(
    code_file_manager: CodeFileManager, code_changes: Iterable[CodeChange]
):
    warned = set()
    for change in code_changes:
        if change.action == CodeChangeAction.CreateFile and os.path.exists(change.file):
            logging.info("Model tried to create a file that already exists")
            cprint(
                f"Error: tried to create {change.file}, but file already existed!",
                color="red",
            )
            raise KeyboardInterrupt
        elif change.action == CodeChangeAction.DeleteFile and not os.path.exists(
            change.file
        ):
            logging.info("Model tried to delete a file that didn't exist")
            cprint(
                f"Error: tried to delete {change.file}, but file doesn't exist!",
                color="red",
            )
            raise KeyboardInterrupt

        if (
            os.path.join(code_file_manager.git_root, change.file)
            not in code_file_manager.file_paths
            and change.action != CodeChangeAction.CreateFile
            and change.file not in warned
        ):
            warned.add(change.file)
            cprint(
                f"Change in {change.file} does not match original files!",
                color="light_yellow",
            )


def get_user_feedback_on_changes(
    config: ConfigManager,
    conv: Conversation,
    user_input_manager: UserInputManager,
    code_file_manager: CodeFileManager,
    code_changes: Iterable[CodeChange],
) -> bool:
    cprint(
        "Apply these changes? 'Y/n/i' or provide feedback.",
        color="light_blue",
    )
    user_response = user_input_manager.collect_user_input()

    need_user_request = True
    match user_response.lower():
        case "y" | "":
            code_changes_to_apply = code_changes
            conv.add_user_message("User chose to apply all your changes.")
        case "n":
            code_changes_to_apply = []
            conv.add_user_message("User chose not to apply any of your changes.")
        case "i":
            code_changes_to_apply, indices = user_filter_changes(
                user_input_manager, code_changes
            )
            conv.add_user_message(
                "User chose to apply"
                f" {len(code_changes_to_apply)}/{len(code_changes)} of your suggest"
                " changes. The changes they applied were:"
                f" {', '.join(map(str, indices))}"
            )
        case _:
            need_user_request = False
            code_changes_to_apply = []
            conv.add_user_message(
                "User chose not to apply any of your changes. User response:"
                f" {user_response}\n\nPlease adjust your previous plan and changes to"
                " reflect this. Respond with a full new set of changes."
            )

    if code_changes_to_apply:
        code_file_manager.write_changes_to_files(code_changes_to_apply)
        if len(code_changes_to_apply) == len(code_changes):
            cprint("Changes applied.", color="light_blue")
        else:
            cprint("Selected changes applied.", color="light_blue")
    else:
        cprint("No changes applied.", color="light_blue")

    if need_user_request:
        cprint("Can I do anything else for you?", color="light_blue")

    return need_user_request


def user_filter_changes(
    user_input_manager: UserInputManager, code_changes: Iterable[CodeChange]
) -> Iterable[CodeChange]:
    new_changes = []
    indices = []
    for index, change in enumerate(code_changes, start=1):
        print_change(change)
        cprint("Keep this change?", "light_blue")
        if user_input_manager.ask_yes_no(default_yes=True):
            new_changes.append(change)
            indices.append(index)
    return new_changes, indices



import os
from enum import Enum

from pygments.lexers import TextLexer, get_lexer_for_filename
from pygments.util import ClassNotFound


class CodeChangeAction(Enum):
    Insert = "insert"
    Replace = "replace"
    Delete = "delete"
    CreateFile = "create-file"
    DeleteFile = "delete-file"

    def has_file(self):
        return self != CodeChangeAction.CreateFile

    def has_surrounding_lines(self):
        return (
            self != CodeChangeAction.CreateFile and self != CodeChangeAction.DeleteFile
        )

    def has_removals(self):
        return (
            self == CodeChangeAction.Delete
            or self == CodeChangeAction.Replace
            or self == CodeChangeAction.DeleteFile
        )

    def has_additions(self):
        return (
            self == CodeChangeAction.Insert
            or self == CodeChangeAction.Replace
            or self == CodeChangeAction.CreateFile
        )


class CodeChange:
    def __init__(
        self,
        json_data: dict,
        code_lines: list[str],
        git_root: str,
        code_file_manager,
    ):
        self.json_data = json_data
        self.code_lines = code_lines
        self.first_changed_line = None
        self.file = self.json_data["file"]
        self.action = CodeChangeAction(self.json_data["action"])
        self.first_changed_line = None
        self.last_changed_line = None
        self.git_root = git_root
        try:
            self.lexer = get_lexer_for_filename(self.file)
            self.lexer.stripnl = False
            self.lexer.stripall = False
            self.lexer.ensurenl = False
        except ClassNotFound:
            self.lexer = TextLexer()

        match self.action:
            case CodeChangeAction.Insert:
                if "insert-before-line" in self.json_data:
                    self.first_changed_line = self.json_data["insert-before-line"]
                    if "insert-after-line" in self.json_data:
                        assert (
                            self.first_changed_line - 1
                            == self.json_data["insert-after-line"]
                        ), (
                            "insert-before-line and insert-after-line are not"
                            " consecutive"
                        )
                elif "insert-after-line" in self.json_data:
                    self.first_changed_line = self.json_data["insert-after-line"] + 1
                else:
                    raise Exception(
                        "insert-before-line or insert-after-line must be specified"
                    )
                self.first_changed_line -= 0.5
                self.last_changed_line = self.first_changed_line

            case CodeChangeAction.Replace:
                self.first_changed_line = self.json_data["start-line"]
                self.last_changed_line = self.json_data["end-line"]

            case CodeChangeAction.Delete:
                self.first_changed_line = self.json_data["start-line"]
                self.last_changed_line = self.json_data["end-line"]

            case CodeChangeAction.CreateFile:
                pass

            case CodeChangeAction.DeleteFile:
                pass

            case _:
                raise Exception(f"Unknown action {self.action}")

        if self.action.has_file():
            file_path = os.path.join(self.git_root, self.file)
            try:
                self.file_lines = code_file_manager.file_lines[file_path]
            except KeyError:
                raise Exception(f"File {file_path} not found!")
            self.line_number_buffer = len(str(len(self.file_lines) + 1)) + 1
        else:
            self.file_lines = []
            self.line_number_buffer = 2

    def __lt__(self, other):
        return self.last_changed_line < other.last_changed_line

    def apply(self, cur_file_lines: list[str]) -> list[str]:
        match self.action:
            case CodeChangeAction.Insert:
                previous_lines = cur_file_lines[: int(self.first_changed_line)]
                following_lines = cur_file_lines[int(self.first_changed_line) :]
                new_file_lines = previous_lines + self.code_lines + following_lines

            case CodeChangeAction.Replace:
                previous_lines = cur_file_lines[: self.first_changed_line - 1]
                following_lines = cur_file_lines[self.last_changed_line :]
                new_file_lines = previous_lines + self.code_lines + following_lines

            case CodeChangeAction.Delete:
                previous_lines = cur_file_lines[: self.first_changed_line - 1]
                following_lines = cur_file_lines[self.last_changed_line :]
                new_file_lines = previous_lines + following_lines

            case CodeChangeAction.CreateFile | CodeChangeAction.DeleteFile:
                raise Exception(
                    f"CodeChange with action={self.action} shouldn't have apply called"
                )

            case _:
                raise Exception(f"Unknown action {self.action.value}")

        return new_file_lines



import asyncio
import json
import logging
from enum import Enum
from timeit import default_timer
from typing import Generator

import attr
import openai
from termcolor import cprint

from .code_change import CodeChange
from .code_change_display import (
    change_delimiter,
    get_file_name,
    get_later_lines,
    get_previous_lines,
    get_removed_block,
)
from .code_file_manager import CodeFileManager
from .llm_api import call_llm_api
from .streaming_printer import StreamingPrinter


class _BlockIndicator(Enum):
    Start = "@@start"
    Code = "@@code"
    End = "@@end"


@attr.s
class ParsingState:
    git_root: str = attr.ib(init=True)
    message: str = attr.ib(default="")
    cur_line: str = attr.ib(default="")
    cur_printed: bool = attr.ib(default=False)
    time_elapsed: float = attr.ib(default=0)
    in_special_lines: bool = attr.ib(default=False)
    in_code_lines: bool = attr.ib(default=False)
    explanation: str = attr.ib(default="")
    explained_since_change: bool = attr.ib(default=True)
    code_changes: list[CodeChange] = attr.ib(factory=list)
    json_lines: list[str] = attr.ib(factory=list)
    code_lines: list[str] = attr.ib(factory=list)

    def parse_line_printing(self, content):
        to_print = ""
        if self.cur_printed:
            to_print = content
        elif not self.could_be_special():
            to_print = self.cur_line
            self.cur_printed = True

        if not self.in_special_lines:
            self.explanation += to_print
            if to_print:
                self.explained_since_change = True

        return to_print

    def could_be_special(self):
        return any(
            len(self.cur_line.rstrip("\n")) <= len(to_match.value)
            and to_match.value.startswith(self.cur_line.rstrip("\n"))
            for to_match in _BlockIndicator
        )

    def create_code_change(self, code_file_manager: CodeFileManager):
        json_data = json.loads("\n".join(self.json_lines))
        self.code_changes.append(
            CodeChange(json_data, self.code_lines, self.git_root, code_file_manager)
        )
        self.json_lines, self.code_lines = [], []

    def new_line(self, code_file_manager: CodeFileManager):
        to_print = ""
        entered_code_lines = False
        exited_code_lines = False
        created_code_change = False
        match self.cur_line.rstrip("\n"):
            case _BlockIndicator.Start.value:
                assert not self.in_special_lines and not self.in_code_lines
                self.in_special_lines = True
            case _BlockIndicator.Code.value:
                assert self.in_special_lines and not self.in_code_lines
                self.in_code_lines = True
                self.create_code_change(code_file_manager)
                entered_code_lines = True
                created_code_change = True
            case _BlockIndicator.End.value:
                assert self.in_special_lines
                if not self.in_code_lines:
                    self.create_code_change(code_file_manager)
                    created_code_change = True
                else:
                    self.code_changes[-1].code_lines = self.code_lines
                    self.code_lines = []
                    exited_code_lines = True
                self.in_special_lines, self.in_code_lines = False, False
            case _:
                if self.in_code_lines:
                    self.code_lines.append(self.cur_line.rstrip("\n"))
                elif self.in_special_lines:
                    self.json_lines.append(self.cur_line)
                elif not self.cur_printed:
                    self.explanation += self.cur_line
                    if self.cur_line:
                        self.explained_since_change = True

                if not self.cur_printed and (
                    self.in_code_lines or not self.in_special_lines
                ):
                    # Lets us print lines that start with @@start
                    to_print = self.cur_line

        self.cur_line = ""
        self.cur_printed = False
        return to_print, entered_code_lines, exited_code_lines, created_code_change


def run_async_stream_and_parse_llm_response(
    messages: list[dict[str, str]],
    model: str,
    code_file_manager: CodeFileManager,
) -> ParsingState:
    state: ParsingState = ParsingState(git_root=code_file_manager.git_root)
    start_time = default_timer()
    try:
        asyncio.run(
            stream_and_parse_llm_response(messages, model, state, code_file_manager)
        )
    except openai.error.InvalidRequestError as e:
        cprint(e, "red")
        exit()
    except KeyboardInterrupt:
        print("\n\nInterrupted by user. Using the response up to this point.")
        # if the last change is incomplete, remove it
        if state.in_code_lines:
            state.code_changes = state.code_changes[:-1]
        logging.info("User interrupted response.")

    state.time_elapsed = default_timer() - start_time
    return state


async def stream_and_parse_llm_response(
    messages: list[dict[str, str]],
    model: str,
    state: ParsingState,
    code_file_manager: CodeFileManager,
) -> None:
    response = await call_llm_api(messages, model)

    print("\nstreaming...  use control-c to interrupt the model at any point\n")

    printer = StreamingPrinter()
    printer_task = asyncio.create_task(printer.print_lines())
    await _process_response(state, response, printer, code_file_manager)
    printer.wrap_it_up()
    await printer_task

    logging.debug(f"LLM response:\n{state.message}")


async def _process_response(
    state: ParsingState,
    response: Generator,
    printer: StreamingPrinter,
    code_file_manager: CodeFileManager,
):
    def chunk_to_lines(chunk):
        return chunk["choices"][0]["delta"].get("content", "").splitlines(keepends=True)

    async for chunk in response:
        for content_line in chunk_to_lines(chunk):
            if content_line:
                state.message += content_line
                _process_content_line(state, content_line, printer, code_file_manager)
    _process_content_line(state, "\n", printer, code_file_manager)


def _process_content_line(
    state, content, printer: StreamingPrinter, code_file_manager: CodeFileManager
):
    beginning = state.cur_line == ""
    state.cur_line += content

    if state.in_code_lines or not state.in_special_lines:
        to_print = state.parse_line_printing(content)
        prefix = (
            "+" + " " * (state.code_changes[-1].line_number_buffer - 1)
            if state.in_code_lines and beginning
            else ""
        )
        color = "green" if state.in_code_lines else None
        if to_print:
            printer.add_string(prefix + to_print, end="", color=color)

    if "\n" in content:
        to_print, entered_code_lines, exited_code_lines, created_code_change = (
            state.new_line(code_file_manager)
        )

        if created_code_change:
            cur_change = state.code_changes[-1]
            if (
                len(state.code_changes) < 2
                or state.code_changes[-2].file != cur_change.file
                or state.explained_since_change
            ):
                printer.add_string(get_file_name(cur_change))
                printer.add_string(change_delimiter)
            state.explained_since_change = False
            printer.add_string(get_previous_lines(cur_change))
            printer.add_string(get_removed_block(cur_change))
            if not cur_change.action.has_additions():
                printer.add_string(get_later_lines(cur_change))
                printer.add_string(change_delimiter)

        prefix = (
            "+" + " " * (state.code_changes[-1].line_number_buffer - 1)
            if state.in_code_lines and beginning
            else ""
        )
        color = "green" if state.in_code_lines else None
        if to_print:
            printer.add_string(prefix + to_print, end="", color=color)
        if exited_code_lines:
            printer.add_string(get_later_lines(state.code_changes[-1]))
            printer.add_string(change_delimiter)



import logging
import os

from prompt_toolkit import PromptSession
from prompt_toolkit.application.current import get_app
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.filters import Condition
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.styles import Style
from termcolor import cprint

from .config_manager import ConfigManager, mentat_dir_path


class FilteredFileHistory(FileHistory):
    def __init__(self, filename: str, config: ConfigManager):
        self.excluded_phrases = ["y", "n", "i", "q"]
        super().__init__(filename)

    def append_string(self, string):
        if string.strip().lower() not in self.excluded_phrases:
            super().append_string(string)


class UserInputManager:
    def __init__(self, config: ConfigManager):
        self.config = config
        self.file_history = FilteredFileHistory(
            os.path.join(mentat_dir_path, "history"), config
        )
        self.auto_suggest = AutoSuggestFromHistory()
        self.style = Style(config.input_style())
        self.prompt = [("class:prompt", ">>> ")]
        self.bindings = KeyBindings()
        self.session = PromptSession(
            message=self.prompt,
            history=self.file_history,
            auto_suggest=self.auto_suggest,
            style=self.style,
            multiline=True,
            prompt_continuation=self.prompt_continuation,
            key_bindings=self.bindings,
        )

        @self.bindings.add("s-down")
        @self.bindings.add("c-j")
        def _(event):
            event.current_buffer.insert_text("\n")

        @self.bindings.add("enter")
        def _(event):
            event.current_buffer.validate_and_handle()

        @Condition
        def suggestion_available() -> bool:
            app = get_app()
            return (
                app.current_buffer.suggestion is not None
                and len(app.current_buffer.suggestion.text) > 0
                and app.current_buffer.document.is_cursor_at_the_end
            )

        # c-i is the code for tab
        @self.bindings.add("c-i", filter=suggestion_available)
        def _(event):
            suggestion = event.current_buffer.suggestion
            if suggestion:
                event.current_buffer.insert_text(suggestion.text)

    def prompt_continuation(self, width, line_number, is_soft_wrap):
        return (
            "" if is_soft_wrap else [("class:continuation", " " * (width - 2) + "> ")]
        )

    def collect_user_input(self) -> str:
        user_input = self.session.prompt().strip()
        logging.debug(f"User input:\n{user_input}")
        if user_input.lower() == "q":
            raise KeyboardInterrupt("User used 'q' to quit")
        return user_input

    def ask_yes_no(self, default_yes: bool) -> bool:
        cprint("(Y/n)" if default_yes else "(y/N)")
        while (user_input := self.collect_user_input().lower()) not in [
            "y",
            "n",
            "",
        ]:
            cprint("(Y/n)" if default_yes else "(y/N)")
        return user_input == "y" or (user_input != "n" and default_yes)



import logging
import os
import sys
from dataclasses import dataclass
from typing import Generator

import openai
import tiktoken
from dotenv import load_dotenv
from termcolor import cprint

from .config_manager import mentat_dir_path, user_config_file_name

package_name = __name__.split(".")[0]


# Check for .env file or already exported API key
# If no api key found, exit and warn user
def setup_api_key():
    if not load_dotenv(os.path.join(mentat_dir_path, ".env")):
        load_dotenv()
    key = os.getenv("OPENAI_API_KEY")
    try:
        openai.api_key = key
        openai.Model.list()  # Test the API key
    except openai.error.AuthenticationError:
        cprint(
            (
                "No valid OpenAI api key detected.\nEither place your key into a .env"
                " file or export it as an environment variable."
            ),
            "red",
        )
        sys.exit(0)


async def call_llm_api(messages: list[dict[str, str]], model) -> Generator:
    if (
        "PYTEST_CURRENT_TEST" in os.environ
        and "--benchmark" not in sys.argv
        and os.getenv("MENTAT_BENCHMARKS_RUNNING") == "false"
    ):
        logging.critical("OpenAI call made in non benchmark test environment!")
        sys.exit(1)

    response = await openai.ChatCompletion.acreate(
        model=model,
        messages=messages,
        temperature=0.5,
        stream=True,
    )

    return response


def count_tokens(message: str) -> int:
    return len(tiktoken.encoding_for_model("gpt-4").encode(message))


def check_model_availability(allow_32k: bool) -> bool:
    available_models = [x["id"] for x in openai.Model.list()["data"]]
    if allow_32k:
        # check if user has access to gpt-4-32k
        if "gpt-4-32k-0314" not in available_models:
            cprint(
                (
                    "You set ALLOW_32K to true, but your OpenAI API key doesn't"
                    " have access to gpt-4-32k-0314. To remove this warning, set"
                    " ALLOW_32K to false until you have access."
                ),
                "yellow",
            )
            allow_32k = False

    if not allow_32k:
        # check if user has access to gpt-4
        if "gpt-4-0314" not in available_models:
            cprint(
                (
                    "Sorry, but your OpenAI API key doesn't have access to gpt-4-0314,"
                    " which is currently required to run Mentat."
                ),
                "red",
            )
            raise KeyboardInterrupt

    return allow_32k


def choose_model(messages: list[dict[str, str]], allow_32k) -> str:
    prompt_token_count = 0
    tokenizer = tiktoken.encoding_for_model("gpt-4")
    for message in messages:
        encoding = tokenizer.encode(message["content"])
        prompt_token_count += len(encoding)
    cprint(f"\nTotal token count: {prompt_token_count}", "cyan")

    model = "gpt-4-0314"
    token_buffer = 500
    if prompt_token_count > 8192 - token_buffer:
        if allow_32k:
            model = "gpt-4-32k-0314"
            if prompt_token_count > 32768 - token_buffer:
                cprint(
                    "Warning: gpt-4-32k-0314 has a token limit of 32768. Attempting"
                    " to run anyway:"
                )
        else:
            cprint(
                (
                    "Warning: gpt-4-0314 has a maximum context length of 8192 tokens."
                    " If you have access to gpt-4-32k-0314, set allow-32k to `true` in"
                    f" `{os.path.join(mentat_dir_path, user_config_file_name)}` to use"
                    " it. Attempting to run with gpt-4-0314:"
                ),
                "yellow",
            )
    return model, prompt_token_count


@dataclass
class CostTracker:
    total_cost: int = 0

    def display_api_call_stats(
        self,
        num_prompt_tokens: int,
        num_sampled_tokens: int,
        model: str,
        call_time: float,
    ) -> None:
        cost_per_1000_tokens = {
            "gpt-4-0314": (0.03, 0.06),
            "gpt-4-32k-0314": (0.06, 0.12),
        }
        prompt_cost = (num_prompt_tokens / 1000) * cost_per_1000_tokens[model][0]
        sampled_cost = (num_sampled_tokens / 1000) * cost_per_1000_tokens[model][1]

        tokens_per_second = num_sampled_tokens / call_time
        call_cost = prompt_cost + sampled_cost

        speed_and_cost_string = (
            f"Speed: {tokens_per_second:.2f} tkns/s | Cost: ${call_cost:.2f}"
        )
        cprint(speed_and_cost_string, "cyan")

        costs_logger = logging.getLogger("costs")
        costs_logger.info(speed_and_cost_string)

        self.total_cost += call_cost

    def display_total_cost(self) -> None:
        cprint(f"\nTotal session cost: ${self.total_cost:.2f}", color="light_blue")



import glob
import logging
import math
import mimetypes
import os
import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Iterable

from termcolor import cprint

from .change_conflict_resolution import (
    resolve_insertion_conflicts,
    resolve_non_insertion_conflicts,
)
from .code_change import CodeChange, CodeChangeAction
from .config_manager import ConfigManager
from .user_input_manager import UserInputManager

# mimetypes is OS dependent, so it's best to add as many extensions as we can
# fmt: off
default_filetype_include_list = [
    ".py",      # Python
    ".java",    # Java
    ".scala",   # Scala
    ".kt",      # Kotlin
    ".php",     # PHP
    ".html",    # HTML
    ".css",     # CSS
    ".less",    # Less
    ".scss",    # SCSS
    ".js",      # Javascript
    ".ts",      # Typescript
    ".c",       # C
    ".cpp",     # C++
    ".h",       # C Header
    ".cs",      # C#
    ".go",      # Go
    ".rs",      # Rust
    ".rb",      # Ruby
    ".swift",   # Swift
    ".lua",     # Lua
    ".pl",      # Perl
    ".sql",     # SQL
    ".r",       # R
    ".m",       # objective-c
    ".sh",      # shell scripts
    ".f",       # fortran
    ".jsx",     # javascript react
    ".tsx",     # typescript react
]
default_filetype_exclude_list = []
# fmt: on


def _get_git_diff_for_path(git_root, path: str) -> str:
    return subprocess.check_output(["git", "diff", path], cwd=git_root).decode("utf-8")


def _get_paths_with_git_diffs(git_root) -> set[str]:
    changed = subprocess.check_output(
        ["git", "diff", "--name-only"], cwd=git_root, text=True
    ).split("\n")
    return set(
        map(lambda path: os.path.realpath(os.path.join(git_root, Path(path))), changed)
    )


def _get_git_root_for_path(path) -> str:
    if os.path.isdir(path):
        dir_path = path
    else:
        dir_path = os.path.dirname(path)
    try:
        git_root = (
            subprocess.check_output(
                [
                    "git",
                    "rev-parse",
                    "--show-toplevel",
                ],
                cwd=os.path.realpath(dir_path),
                stderr=subprocess.DEVNULL,
            )
            .decode("utf-8")
            .strip()
        )
        # call realpath to resolve symlinks, so all paths match
        return os.path.realpath(git_root)
    except subprocess.CalledProcessError:
        logging.error(f"File {path} isn't part of a git project.")
        exit()


def _get_shared_git_root_for_paths(paths) -> str:
    git_roots = set()
    for path in paths:
        git_root = _get_git_root_for_path(path)
        git_roots.add(git_root)
    if not paths:
        git_root = _get_git_root_for_path(os.getcwd())
        git_roots.add(git_root)

    if len(git_roots) > 1:
        logging.error(
            "All paths must be part of the same git project! Projects provided:"
            f" {git_roots}"
        )
        exit()
    elif len(git_roots) == 0:
        logging.error("No git projects provided.")
        exit()

    return git_roots.pop()


def _build_path_tree(file_paths, git_root):
    tree = {}
    for path in file_paths:
        path = os.path.relpath(path, git_root)
        parts = Path(path).parts
        current_level = tree
        for part in parts:
            if part not in current_level:
                current_level[part] = {}
            current_level = current_level[part]
    return tree


def _print_path_tree(tree, non_text_files, changed_files, cur_path, prefix=""):
    keys = list(tree.keys())
    for i, key in enumerate(sorted(keys)):
        if i < len(keys) - 1:
            new_prefix = prefix + "│   "
            print(f"{prefix}├── ", end="")
        else:
            new_prefix = prefix + "    "
            print(f"{prefix}└── ", end="")

        cur = os.path.join(cur_path, key)
        star = "* " if cur in changed_files else ""
        if tree[key]:
            color = "blue"
        elif cur in non_text_files:
            color = "yellow"
        elif star:
            color = "green"
        else:
            color = None
        cprint(f"{star}{key}", color)
        if tree[key]:
            _print_path_tree(tree[key], non_text_files, changed_files, cur, new_prefix)


def _is_file_text(file_name):
    file_type, encoding = mimetypes.guess_type(file_name)
    return file_type and file_type.split("/")[0] == "text"


class CodeFileManager:
    def __init__(
        self,
        paths: Iterable[str],
        user_input_manager: UserInputManager,
        config: ConfigManager,
    ):
        # Make sure to apply user config last
        for file_type in default_filetype_include_list:
            mimetypes.add_type("text/default-include-list", file_type)
        for file_type in default_filetype_exclude_list:
            mimetypes.types_map.pop(file_type, None)

        for file_type in config.filetype_include_list():
            mimetypes.add_type("text/user-include-list", file_type)
        for file_type in config.filetype_exclude_list():
            mimetypes.types_map.pop(file_type, None)

        self.config = config
        self.git_root = _get_shared_git_root_for_paths(paths)
        self._set_file_paths(paths)
        self.user_input_manager = user_input_manager

        if self.file_paths:
            cprint("Files included in context:", "green")
        else:
            cprint("No files included in context.\n", "red")
            cprint("Git project: ", "green", end="")
        cprint(os.path.split(self.git_root)[1], "blue")
        _print_path_tree(
            _build_path_tree(self.file_paths + self.non_text_file_paths, self.git_root),
            self.non_text_file_paths,
            _get_paths_with_git_diffs(self.git_root),
            self.git_root,
        )

    def _set_file_paths(self, paths: Iterable[str] = None) -> None:
        invalid_paths = []
        for path in paths:
            if not os.path.exists(path):
                invalid_paths.append(path)
        if invalid_paths:
            cprint("Error:", "red", end=" ")
            cprint("The following paths do not exist:")
            print("\n".join(invalid_paths))
            print("Exiting...")
            exit()

        path_set = set()
        self.non_text_file_paths = []
        self.file_paths = []

        for path in paths:
            path = Path(path)
            if path.is_file():
                path_set.add(os.path.realpath(path))
            elif path.is_dir():
                non_git_ignored_files = set(
                    # git returns / seperated paths even on windows, convert so we can remove
                    # glob_excluded_files, which have windows paths on windows
                    os.path.normpath(path)
                    for path in filter(
                        lambda p: p != "",
                        subprocess.check_output(
                            # -c shows cached (regular) files, -o shows other (untracked/ new) files
                            ["git", "ls-files", "-c", "-o", "--exclude-standard"],
                            cwd=path,
                            text=True,
                        ).split("\n"),
                    )
                )
                glob_excluded_files = set(
                    file
                    for glob_path in self.config.file_exclude_glob_list()
                    # If the user puts a / at the beginning, it will try to look in root directory
                    for file in glob.glob(
                        pathname=glob_path.lstrip("/"),
                        root_dir=path,
                        recursive=True,
                    )
                )
                nonignored_files = non_git_ignored_files - glob_excluded_files

                non_text_files = filter(
                    lambda f: not _is_file_text(
                        os.path.realpath(os.path.join(path, f))
                    ),
                    nonignored_files,
                )
                self.non_text_file_paths.extend(
                    map(
                        lambda f: os.path.realpath(os.path.join(path, f)),
                        non_text_files,
                    )
                )
                text_files = filter(
                    _is_file_text,
                    map(
                        lambda f: os.path.realpath(os.path.join(path, f)),
                        nonignored_files,
                    ),
                )
                path_set.update(text_files)
        self.file_paths = list(path_set)

    def _read_file(self, abs_path) -> Iterable[str]:
        with open(abs_path, "r") as f:
            lines = f.read().split("\n")
        return lines

    def _read_all_file_lines(self) -> None:
        self.file_lines = dict()
        for abs_path in self.file_paths:
            self.file_lines[abs_path] = self._read_file(abs_path)

    def get_code_message(self):
        self._read_all_file_lines()
        code_message = ["Code Files:\n"]
        for abs_path in self.file_paths:
            rel_path = os.path.relpath(abs_path, self.git_root)
            code_message.append(rel_path)
            for i, line in enumerate(self.file_lines[abs_path], start=1):
                code_message.append(f"{i}:{line}")
            code_message.append("")
            git_diff_output = _get_git_diff_for_path(self.git_root, rel_path)
            if git_diff_output:
                code_message.append("Current git diff for this file:")
                code_message.append(f"{git_diff_output}")
        if self.non_text_file_paths:
            code_message.append("\nOther files:\n")
            code_message.extend(
                os.path.relpath(path, self.git_root)
                for path in self.non_text_file_paths
            )
        return "\n".join(code_message)

    def _handle_delete(self, delete_change):
        file_path = os.path.join(self.git_root, delete_change.file)
        if not os.path.exists(file_path):
            logging.error(f"Path {file_path} non-existent on delete")
            return

        cprint(f"Are you sure you want to delete {delete_change.file}?", "red")
        if self.user_input_manager.ask_yes_no(default_yes=False):
            logging.info(f"Deleting file {file_path}")
            cprint(f"Deleting {delete_change.file}...")
            self.file_paths.remove(file_path)
            os.remove(file_path)
        else:
            cprint(f"Not deleting {delete_change.file}")

    def _get_new_code_lines(self, changes) -> Iterable[str]:
        if len(set(map(lambda change: change.file, changes))) > 1:
            raise Exception("All changes passed in must be for the same file")

        changes = sorted(changes, reverse=True)

        # We resolve insertion conflicts twice because non-insertion conflicts
        # might move insert blocks outside of replace/delete blocks and cause
        # them to conflict again
        changes = resolve_insertion_conflicts(changes, self.user_input_manager, self)
        changes = resolve_non_insertion_conflicts(changes, self.user_input_manager)
        changes = resolve_insertion_conflicts(changes, self.user_input_manager, self)
        if not changes:
            return []

        rel_path = changes[0].file
        abs_path = os.path.join(self.git_root, rel_path)
        new_code_lines = self.file_lines[abs_path].copy()
        if new_code_lines != self._read_file(abs_path):
            logging.info(f"File '{rel_path}' changed while generating changes")
            cprint(
                (
                    f"File '{rel_path}' changed while generating; current file changes"
                    " will be erased. Continue?"
                ),
                color="light_yellow",
            )
            if not self.user_input_manager.ask_yes_no(default_yes=False):
                cprint(f"Not applying changes to file {rel_path}.")
                return None

        # Necessary in case the model needs to insert past the end of the file
        last_line = len(new_code_lines) + 1
        largest_changed_line = math.ceil(changes[0].last_changed_line)
        if largest_changed_line > last_line:
            new_code_lines += [""] * (largest_changed_line - last_line)

        min_changed_line = largest_changed_line + 1
        for i, change in enumerate(changes):
            if change.last_changed_line >= min_changed_line:
                raise ValueError(f"Change line number overlap in file {change.file}")
            min_changed_line = change.first_changed_line
            new_code_lines = change.apply(new_code_lines)
        return new_code_lines

    def write_changes_to_files(self, code_changes: list[CodeChange]) -> None:
        files_to_write = dict()
        file_changes = defaultdict(list)
        for code_change in code_changes:
            rel_path = code_change.file
            if code_change.action == CodeChangeAction.CreateFile:
                cprint(f"Creating new file {rel_path}", color="light_green")
                files_to_write[rel_path] = code_change.code_lines
            elif code_change.action == CodeChangeAction.DeleteFile:
                self._handle_delete(code_change)
            else:
                file_changes[rel_path].append(code_change)

        for file_path, changes in file_changes.items():
            new_code_lines = self._get_new_code_lines(changes)
            if new_code_lines:
                files_to_write[file_path] = new_code_lines

        for rel_path, code_lines in files_to_write.items():
            file_path = os.path.join(self.git_root, rel_path)
            # if changes created a new file and we are now writing to it, add it to context
            if file_path not in self.file_paths:
                logging.info(f"Adding new file {file_path} to context")
                self.file_paths.append(file_path)
            with open(file_path, "w") as f:
                f.write("\n".join(code_lines))



from setuptools import find_packages, setup


def read_requirements(file):
    with open(file, "r") as f:
        return f.read().splitlines()


setup(
    name="mentat-ai",
    version="0.1.4",
    python_requires=">=3.10",
    packages=find_packages(),
    install_requires=read_requirements("requirements.txt"),
    package_data={
        "mentat": ["default_config.json"],
    },
    entry_points={
        "console_scripts": [
            "mentat=mentat.app:run_cli",
        ],
    },
)



import os
import subprocess
from unittest import TestCase

import pytest

from mentat.code_file_manager import CodeFileManager
from mentat.config_manager import ConfigManager

config = ConfigManager()


def test_path_gitignoring(temp_testbed):
    gitignore_path = ".gitignore"
    testing_dir_path = "git_testing_dir"
    os.makedirs(testing_dir_path)

    # create 3 files, 2 ignored in gitignore, 1 not
    ignored_file_path_1 = os.path.join(testing_dir_path, "ignored_file_1.txt")
    ignored_file_path_2 = os.path.join(testing_dir_path, "ignored_file_2.txt")
    non_ignored_file_path = os.path.join(testing_dir_path, "non_ignored_file.txt")

    with open(gitignore_path, "w") as gitignore_file:
        gitignore_file.write("ignored_file_1.txt\nignored_file_2.txt")

    for file_path in [ignored_file_path_1, ignored_file_path_2, non_ignored_file_path]:
        with open(file_path, "w") as file:
            file.write("I am a file")

    # Run CodeFileManager on the git_testing_dir, and also explicitly pass in ignored_file_2.txt
    paths = [testing_dir_path, ignored_file_path_2]
    code_file_manager = CodeFileManager(paths, user_input_manager=None, config=config)

    expected_file_paths = [
        os.path.join(temp_testbed, ignored_file_path_2),
        os.path.join(temp_testbed, non_ignored_file_path),
    ]

    case = TestCase()
    case.assertListEqual(
        sorted(expected_file_paths), sorted(code_file_manager.file_paths)
    )


def test_ignore_non_text_files(temp_testbed):
    image_file_path = "image.jpg"
    with open(image_file_path, "w") as image_file:
        image_file.write("I am an image")
    code_file_manager = CodeFileManager(["./"], user_input_manager=None, config=config)
    assert (
        os.path.join(temp_testbed, image_file_path) not in code_file_manager.file_paths
    )


def test_no_paths_given(temp_testbed):
    # Get temp_testbed as the git root when given no paths
    code_file_manager = CodeFileManager([], user_input_manager=None, config=config)
    assert code_file_manager.git_root == temp_testbed


def test_paths_given(temp_testbed):
    # Get temp_testbed when given file in temp_testbed
    code_file_manager = CodeFileManager(
        ["scripts"], user_input_manager=None, config=config
    )
    assert code_file_manager.git_root == temp_testbed


def test_two_git_roots_given():
    # Exits when given 2 paths with separate git roots
    with pytest.raises(SystemExit) as e_info:
        os.makedirs("git_testing_dir")
        subprocess.run(["git", "init"], cwd="git_testing_dir")

        _ = CodeFileManager(
            ["./", "git_testing_dir"], user_input_manager=None, config=config
        )
    assert e_info.type == SystemExit


def test_glob_exclude(mocker, temp_testbed):
    # Makes sure glob exclude config works
    mock_glob_exclude = mocker.MagicMock()
    mocker.patch.object(ConfigManager, "file_exclude_glob_list", new=mock_glob_exclude)
    mock_glob_exclude.side_effect = [[os.path.join("glob_test", "**", "*.py")]]

    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.py")
    glob_include_path = os.path.join("glob_test", "bagel", "apple", "include_me.ts")
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am excluded")
    os.makedirs(os.path.dirname(glob_include_path), exist_ok=True)
    with open(glob_include_path, "w") as glob_include_file:
        glob_include_file.write("I am included")

    code_file_manager = CodeFileManager(["."], user_input_manager=None, config=config)
    print(code_file_manager.file_paths)
    assert (
        os.path.join(temp_testbed, glob_exclude_path)
        not in code_file_manager.file_paths
    )
    assert os.path.join(temp_testbed, glob_include_path) in code_file_manager.file_paths
